"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

from dataclasses import dataclass
import asyncio
from typing import Callable

from . import DataClasses, Token
from .MessageClasses import StageLinQServiceAnnouncement, StageLinQMessage
from .DataClasses import StageLinQServiceAnnouncementData
import json

prime_go = {
    "ClientLibrarianDevicesControllerCurrentDevice": "/Client/Librarian/DevicesController/CurrentDevice",
    "ClientLibrarianDevicesControllerHasSDCardConnected": "/Client/Librarian/DevicesController/HasSDCardConnected",
    "ClientLibrarianDevicesControllerHasUsbDeviceConnected": "/Client/Librarian/DevicesController/HasUsbDeviceConnected",
    "ClientPreferencesLayerB": "/Client/Preferences/LayerB",
    "ClientPreferencesPlayer": "/Client/Preferences/Player",
    "ClientPreferencesProfileApplicationPlayerColor1": "/Client/Preferences/Profile/Application/PlayerColor1",
    "ClientPreferencesProfileApplicationPlayerColor1A": "/Client/Preferences/Profile/Application/PlayerColor1A",
    "ClientPreferencesProfileApplicationPlayerColor1B": "/Client/Preferences/Profile/Application/PlayerColor1B",
    "ClientPreferencesProfileApplicationPlayerColor2": "/Client/Preferences/Profile/Application/PlayerColor2",
    "ClientPreferencesProfileApplicationPlayerColor2A": "/Client/Preferences/Profile/Application/PlayerColor2A",
    "ClientPreferencesProfileApplicationPlayerColor2B": "/Client/Preferences/Profile/Application/PlayerColor2B",
    "ClientPreferencesProfileApplicationPlayerColor3A": "/Client/Preferences/Profile/Application/PlayerColor3A",
    "ClientPreferencesProfileApplicationPlayerColor3B": "/Client/Preferences/Profile/Application/PlayerColor3B",
    "ClientPreferencesProfileApplicationPlayerColor4A": "/Client/Preferences/Profile/Application/PlayerColor4A",
    "ClientPreferencesProfileApplicationPlayerColor4B": "/Client/Preferences/Profile/Application/PlayerColor4B",
    "ClientPreferencesProfileApplicationSyncMode": "/Client/Preferences/Profile/Application/SyncMode",
    "EngineDeck1CurrentBPM": "/Engine/Deck1/CurrentBPM",
    "EngineDeck1ExternalMixerVolume": "/Engine/Deck1/ExternalMixerVolume",
    "EngineDeck1ExternalScratchWheelTouch": "/Engine/Deck1/ExternalScratchWheelTouch",
    "EngineDeck1PadsView": "/Engine/Deck1/Pads/View",
    "EngineDeck1Play": "/Engine/Deck1/Play",
    "EngineDeck1PlayState": "/Engine/Deck1/PlayState",
    "EngineDeck1Speed": "/Engine/Deck1/Speed",
    "EngineDeck1SpeedNeutral": "/Engine/Deck1/SpeedNeutral",
    "EngineDeck1SpeedOffsetDown": "/Engine/Deck1/SpeedOffsetDown",
    "EngineDeck1SpeedOffsetUp": "/Engine/Deck1/SpeedOffsetUp",
    "EngineDeck1SpeedRange": "/Engine/Deck1/SpeedRange",
    "EngineDeck1SpeedState": "/Engine/Deck1/SpeedState",
    "EngineDeck1SyncMode": "/Engine/Deck1/SyncMode",
    "EngineDeck1TrackArtistName": "/Engine/Deck1/Track/ArtistName",
    "EngineDeck1TrackBleep": "/Engine/Deck1/Track/Bleep",
    "EngineDeck1TrackCuePosition": "/Engine/Deck1/Track/CuePosition",
    "EngineDeck1TrackCurrentBPM": "/Engine/Deck1/Track/CurrentBPM",
    "EngineDeck1TrackCurrentKeyIndex": "/Engine/Deck1/Track/CurrentKeyIndex",
    "EngineDeck1TrackCurrentLoopInPosition": "/Engine/Deck1/Track/CurrentLoopInPosition",
    "EngineDeck1TrackCurrentLoopOutPosition": "/Engine/Deck1/Track/CurrentLoopOutPosition",
    "EngineDeck1TrackCurrentLoopSizeInBeats": "/Engine/Deck1/Track/CurrentLoopSizeInBeats",
    "EngineDeck1TrackKeyLock": "/Engine/Deck1/Track/KeyLock",
    "EngineDeck1TrackLoopEnableState": "/Engine/Deck1/Track/LoopEnableState",
    "EngineDeck1TrackLoopQuickLoop1": "/Engine/Deck1/Track/Loop/QuickLoop1",
    "EngineDeck1TrackLoopQuickLoop2": "/Engine/Deck1/Track/Loop/QuickLoop2",
    "EngineDeck1TrackLoopQuickLoop3": "/Engine/Deck1/Track/Loop/QuickLoop3",
    "EngineDeck1TrackLoopQuickLoop4": "/Engine/Deck1/Track/Loop/QuickLoop4",
    "EngineDeck1TrackLoopQuickLoop5": "/Engine/Deck1/Track/Loop/QuickLoop5",
    "EngineDeck1TrackLoopQuickLoop6": "/Engine/Deck1/Track/Loop/QuickLoop6",
    "EngineDeck1TrackLoopQuickLoop7": "/Engine/Deck1/Track/Loop/QuickLoop7",
    "EngineDeck1TrackLoopQuickLoop8": "/Engine/Deck1/Track/Loop/QuickLoop8",
    "EngineDeck1TrackPlayPauseLEDState": "/Engine/Deck1/Track/PlayPauseLEDState",
    "EngineDeck1TrackSampleRate": "/Engine/Deck1/Track/SampleRate",
    "EngineDeck1TrackSongAnalyzed": "/Engine/Deck1/Track/SongAnalyzed",
    "EngineDeck1TrackSongLoaded": "/Engine/Deck1/Track/SongLoaded",
    "EngineDeck1TrackSongName": "/Engine/Deck1/Track/SongName",
    "EngineDeck1TrackSoundSwitchGUID": "/Engine/Deck1/Track/SoundSwitchGuid",
    "EngineDeck1TrackTrackBytes": "/Engine/Deck1/Track/TrackBytes",
    "EngineDeck1TrackTrackData": "/Engine/Deck1/Track/TrackData",
    "EngineDeck1TrackTrackLength": "/Engine/Deck1/Track/TrackLength",
    "EngineDeck1TrackTrackName": "/Engine/Deck1/Track/TrackName",
    "EngineDeck1TrackTrackNetworkPath": "/Engine/Deck1/Track/TrackNetworkPath",
    "EngineDeck1TrackTrackURI": "/Engine/Deck1/Track/TrackUri",
    "EngineDeck1TrackTrackWasPlayed": "/Engine/Deck1/Track/TrackWasPlayed",
    "EngineDeck2CurrentBPM": "/Engine/Deck2/CurrentBPM",
    "EngineDeck2ExternalMixerVolume": "/Engine/Deck2/ExternalMixerVolume",
    "EngineDeck2ExternalScratchWheelTouch": "/Engine/Deck2/ExternalScratchWheelTouch",
    "EngineDeck2PadsView": "/Engine/Deck2/Pads/View",
    "EngineDeck2Play": "/Engine/Deck2/Play",
    "EngineDeck2PlayState": "/Engine/Deck2/PlayState",
    "EngineDeck2Speed": "/Engine/Deck2/Speed",
    "EngineDeck2SpeedNeutral": "/Engine/Deck2/SpeedNeutral",
    "EngineDeck2SpeedOffsetDown": "/Engine/Deck2/SpeedOffsetDown",
    "EngineDeck2SpeedOffsetUp": "/Engine/Deck2/SpeedOffsetUp",
    "EngineDeck2SpeedRange": "/Engine/Deck2/SpeedRange",
    "EngineDeck2SpeedState": "/Engine/Deck2/SpeedState",
    "EngineDeck2SyncMode": "/Engine/Deck2/SyncMode",
    "EngineDeck2TrackArtistName": "/Engine/Deck2/Track/ArtistName",
    "EngineDeck2TrackBleep": "/Engine/Deck2/Track/Bleep",
    "EngineDeck2TrackCuePosition": "/Engine/Deck2/Track/CuePosition",
    "EngineDeck2TrackCurrentBPM": "/Engine/Deck2/Track/CurrentBPM",
    "EngineDeck2TrackCurrentKeyIndex": "/Engine/Deck2/Track/CurrentKeyIndex",
    "EngineDeck2TrackCurrentLoopInPosition": "/Engine/Deck2/Track/CurrentLoopInPosition",
    "EngineDeck2TrackCurrentLoopOutPosition": "/Engine/Deck2/Track/CurrentLoopOutPosition",
    "EngineDeck2TrackCurrentLoopSizeInBeats": "/Engine/Deck2/Track/CurrentLoopSizeInBeats",
    "EngineDeck2TrackKeyLock": "/Engine/Deck2/Track/KeyLock",
    "EngineDeck2TrackLoopEnableState": "/Engine/Deck2/Track/LoopEnableState",
    "EngineDeck2TrackLoopQuickLoop1": "/Engine/Deck2/Track/Loop/QuickLoop1",
    "EngineDeck2TrackLoopQuickLoop2": "/Engine/Deck2/Track/Loop/QuickLoop2",
    "EngineDeck2TrackLoopQuickLoop3": "/Engine/Deck2/Track/Loop/QuickLoop3",
    "EngineDeck2TrackLoopQuickLoop4": "/Engine/Deck2/Track/Loop/QuickLoop4",
    "EngineDeck2TrackLoopQuickLoop5": "/Engine/Deck2/Track/Loop/QuickLoop5",
    "EngineDeck2TrackLoopQuickLoop6": "/Engine/Deck2/Track/Loop/QuickLoop6",
    "EngineDeck2TrackLoopQuickLoop7": "/Engine/Deck2/Track/Loop/QuickLoop7",
    "EngineDeck2TrackLoopQuickLoop8": "/Engine/Deck2/Track/Loop/QuickLoop8",
    "EngineDeck2TrackPlayPauseLEDState": "/Engine/Deck2/Track/PlayPauseLEDState",
    "EngineDeck2TrackSampleRate": "/Engine/Deck2/Track/SampleRate",
    "EngineDeck2TrackSongAnalyzed": "/Engine/Deck2/Track/SongAnalyzed",
    "EngineDeck2TrackSongLoaded": "/Engine/Deck2/Track/SongLoaded",
    "EngineDeck2TrackSongName": "/Engine/Deck2/Track/SongName",
    "EngineDeck2TrackSoundSwitchGUID": "/Engine/Deck2/Track/SoundSwitchGuid",
    "EngineDeck2TrackTrackBytes": "/Engine/Deck2/Track/TrackBytes",
    "EngineDeck2TrackTrackData": "/Engine/Deck2/Track/TrackData",
    "EngineDeck2TrackTrackLength": "/Engine/Deck2/Track/TrackLength",
    "EngineDeck2TrackTrackName": "/Engine/Deck2/Track/TrackName",
    "EngineDeck2TrackTrackNetworkPath": "/Engine/Deck2/Track/TrackNetworkPath",
    "EngineDeck2TrackTrackURI": "/Engine/Deck2/Track/TrackUri",
    "EngineDeck2TrackTrackWasPlayed": "/Engine/Deck2/Track/TrackWasPlayed",
    "EngineDeck3CurrentBPM" "EngineDeckCount": "/Engine/DeckCount",
    "MixerCH1faderPosition": "/Mixer/CH1faderPosition",
    "MixerCH2faderPosition": "/Mixer/CH2faderPosition",
    "MixerCH3faderPosition": "/Mixer/CH3faderPosition",
    "MixerCH4faderPosition": "/Mixer/CH4faderPosition",
    "MixerCrossfaderPosition": "/Mixer/CrossfaderPosition",
}

common_functions = {
    "ClientLibrarianDevicesControllerCurrentDevice": "/Client/Librarian/DevicesController/CurrentDevice",
    "ClientLibrarianDevicesControllerHasSDCardConnected": "/Client/Librarian/DevicesController/HasSDCardConnected",
    "ClientLibrarianDevicesControllerHasUsbDeviceConnected": "/Client/Librarian/DevicesController/HasUsbDeviceConnected",
    "ClientPreferencesLayerB": "/Client/Preferences/LayerB",
    "ClientPreferencesPlayer": "/Client/Preferences/Player",
    "ClientPreferencesProfileApplicationPlayerColor1": "/Client/Preferences/Profile/Application/PlayerColor1",
    "ClientPreferencesProfileApplicationPlayerColor1A": "/Client/Preferences/Profile/Application/PlayerColor1A",
    "ClientPreferencesProfileApplicationPlayerColor1B": "/Client/Preferences/Profile/Application/PlayerColor1B",
    "ClientPreferencesProfileApplicationPlayerColor2": "/Client/Preferences/Profile/Application/PlayerColor2",
    "ClientPreferencesProfileApplicationPlayerColor2A": "/Client/Preferences/Profile/Application/PlayerColor2A",
    "ClientPreferencesProfileApplicationPlayerColor2B": "/Client/Preferences/Profile/Application/PlayerColor2B",
    "ClientPreferencesProfileApplicationPlayerColor3": "/Client/Preferences/Profile/Application/PlayerColor3",
    "ClientPreferencesProfileApplicationPlayerColor3A": "/Client/Preferences/Profile/Application/PlayerColor3A",
    "ClientPreferencesProfileApplicationPlayerColor3B": "/Client/Preferences/Profile/Application/PlayerColor3B",
    "ClientPreferencesProfileApplicationPlayerColor4": "/Client/Preferences/Profile/Application/PlayerColor4",
    "ClientPreferencesProfileApplicationPlayerColor4A": "/Client/Preferences/Profile/Application/PlayerColor4A",
    "ClientPreferencesProfileApplicationPlayerColor4B": "/Client/Preferences/Profile/Application/PlayerColor4B",
    "ClientPreferencesProfileApplicationSyncMode": "/Client/Preferences/Profile/Application/SyncMode",
    "EngineDeck1CurrentBPM": "/Engine/Deck1/CurrentBPM",
    "EngineDeck1ExternalMixerVolume": "/Engine/Deck1/ExternalMixerVolume",
    "EngineDeck1ExternalScratchWheelTouch": "/Engine/Deck1/ExternalScratchWheelTouch",
    "EngineDeck1PadsView": "/Engine/Deck1/Pads/View",
    "EngineDeck1Play": "/Engine/Deck1/Play",
    "EngineDeck1PlayState": "/Engine/Deck1/PlayState",
    "EngineDeck1PlayStatePath": "/Engine/Deck1/PlayStatePath",
    "EngineDeck1Speed": "/Engine/Deck1/Speed",
    "EngineDeck1SpeedNeutral": "/Engine/Deck1/SpeedNeutral",
    "EngineDeck1SpeedOffsetDown": "/Engine/Deck1/SpeedOffsetDown",
    "EngineDeck1SpeedOffsetUp": "/Engine/Deck1/SpeedOffsetUp",
    "EngineDeck1SpeedRange": "/Engine/Deck1/SpeedRange",
    "EngineDeck1SpeedState": "/Engine/Deck1/SpeedState",
    "EngineDeck1SyncMode": "/Engine/Deck1/SyncMode",
    "EngineDeck1TrackArtistName": "/Engine/Deck1/Track/ArtistName",
    "EngineDeck1TrackBleep": "/Engine/Deck1/Track/Bleep",
    "EngineDeck1TrackCuePosition": "/Engine/Deck1/Track/CuePosition",
    "EngineDeck1TrackCurrentBPM": "/Engine/Deck1/Track/CurrentBPM",
    "EngineDeck1TrackCurrentKeyIndex": "/Engine/Deck1/Track/CurrentKeyIndex",
    "EngineDeck1TrackCurrentLoopInPosition": "/Engine/Deck1/Track/CurrentLoopInPosition",
    "EngineDeck1TrackCurrentLoopOutPosition": "/Engine/Deck1/Track/CurrentLoopOutPosition",
    "EngineDeck1TrackCurrentLoopSizeInBeats": "/Engine/Deck1/Track/CurrentLoopSizeInBeats",
    "EngineDeck1TrackKeyLock": "/Engine/Deck1/Track/KeyLock",
    "EngineDeck1TrackLoopEnableState": "/Engine/Deck1/Track/LoopEnableState",
    "EngineDeck1TrackLoopQuickLoop1": "/Engine/Deck1/Track/Loop/QuickLoop1",
    "EngineDeck1TrackLoopQuickLoop2": "/Engine/Deck1/Track/Loop/QuickLoop2",
    "EngineDeck1TrackLoopQuickLoop3": "/Engine/Deck1/Track/Loop/QuickLoop3",
    "EngineDeck1TrackLoopQuickLoop4": "/Engine/Deck1/Track/Loop/QuickLoop4",
    "EngineDeck1TrackLoopQuickLoop5": "/Engine/Deck1/Track/Loop/QuickLoop5",
    "EngineDeck1TrackLoopQuickLoop6": "/Engine/Deck1/Track/Loop/QuickLoop6",
    "EngineDeck1TrackLoopQuickLoop7": "/Engine/Deck1/Track/Loop/QuickLoop7",
    "EngineDeck1TrackLoopQuickLoop8": "/Engine/Deck1/Track/Loop/QuickLoop8",
    "EngineDeck1TrackPlayPauseLEDState": "/Engine/Deck1/Track/PlayPauseLEDState",
    "EngineDeck1TrackSampleRate": "/Engine/Deck1/Track/SampleRate",
    "EngineDeck1TrackSongAnalyzed": "/Engine/Deck1/Track/SongAnalyzed",
    "EngineDeck1TrackSongLoaded": "/Engine/Deck1/Track/SongLoaded",
    "EngineDeck1TrackSongName": "/Engine/Deck1/Track/SongName",
    "EngineDeck1TrackSoundSwitchGUID": "/Engine/Deck1/Track/SoundSwitchGuid",
    "EngineDeck1TrackTrackBytes": "/Engine/Deck1/Track/TrackBytes",
    "EngineDeck1TrackTrackData": "/Engine/Deck1/Track/TrackData",
    "EngineDeck1TrackTrackLength": "/Engine/Deck1/Track/TrackLength",
    "EngineDeck1TrackTrackName": "/Engine/Deck1/Track/TrackName",
    "EngineDeck1TrackTrackNetworkPath": "/Engine/Deck1/Track/TrackNetworkPath",
    "EngineDeck1TrackTrackURI": "/Engine/Deck1/Track/TrackUri",
    "EngineDeck1TrackTrackWasPlayed": "/Engine/Deck1/Track/TrackWasPlayed",
    "EngineDeck2CurrentBPM": "/Engine/Deck2/CurrentBPM",
    "EngineDeck2ExternalMixerVolume": "/Engine/Deck2/ExternalMixerVolume",
    "EngineDeck2ExternalScratchWheelTouch": "/Engine/Deck2/ExternalScratchWheelTouch",
    "EngineDeck2PadsView": "/Engine/Deck2/Pads/View",
    "EngineDeck2Play": "/Engine/Deck2/Play",
    "EngineDeck2PlayState": "/Engine/Deck2/PlayState",
    "EngineDeck2PlayStatePath": "/Engine/Deck2/PlayStatePath",
    "EngineDeck2Speed": "/Engine/Deck2/Speed",
    "EngineDeck2SpeedNeutral": "/Engine/Deck2/SpeedNeutral",
    "EngineDeck2SpeedOffsetDown": "/Engine/Deck2/SpeedOffsetDown",
    "EngineDeck2SpeedOffsetUp": "/Engine/Deck2/SpeedOffsetUp",
    "EngineDeck2SpeedRange": "/Engine/Deck2/SpeedRange",
    "EngineDeck2SpeedState": "/Engine/Deck2/SpeedState",
    "EngineDeck2SyncMode": "/Engine/Deck2/SyncMode",
    "EngineDeck2TrackArtistName": "/Engine/Deck2/Track/ArtistName",
    "EngineDeck2TrackBleep": "/Engine/Deck2/Track/Bleep",
    "EngineDeck2TrackCuePosition": "/Engine/Deck2/Track/CuePosition",
    "EngineDeck2TrackCurrentBPM": "/Engine/Deck2/Track/CurrentBPM",
    "EngineDeck2TrackCurrentKeyIndex": "/Engine/Deck2/Track/CurrentKeyIndex",
    "EngineDeck2TrackCurrentLoopInPosition": "/Engine/Deck2/Track/CurrentLoopInPosition",
    "EngineDeck2TrackCurrentLoopOutPosition": "/Engine/Deck2/Track/CurrentLoopOutPosition",
    "EngineDeck2TrackCurrentLoopSizeInBeats": "/Engine/Deck2/Track/CurrentLoopSizeInBeats",
    "EngineDeck2TrackKeyLock": "/Engine/Deck2/Track/KeyLock",
    "EngineDeck2TrackLoopEnableState": "/Engine/Deck2/Track/LoopEnableState",
    "EngineDeck2TrackLoopQuickLoop1": "/Engine/Deck2/Track/Loop/QuickLoop1",
    "EngineDeck2TrackLoopQuickLoop2": "/Engine/Deck2/Track/Loop/QuickLoop2",
    "EngineDeck2TrackLoopQuickLoop3": "/Engine/Deck2/Track/Loop/QuickLoop3",
    "EngineDeck2TrackLoopQuickLoop4": "/Engine/Deck2/Track/Loop/QuickLoop4",
    "EngineDeck2TrackLoopQuickLoop5": "/Engine/Deck2/Track/Loop/QuickLoop5",
    "EngineDeck2TrackLoopQuickLoop6": "/Engine/Deck2/Track/Loop/QuickLoop6",
    "EngineDeck2TrackLoopQuickLoop7": "/Engine/Deck2/Track/Loop/QuickLoop7",
    "EngineDeck2TrackLoopQuickLoop8": "/Engine/Deck2/Track/Loop/QuickLoop8",
    "EngineDeck2TrackPlayPauseLEDState": "/Engine/Deck2/Track/PlayPauseLEDState",
    "EngineDeck2TrackSampleRate": "/Engine/Deck2/Track/SampleRate",
    "EngineDeck2TrackSongAnalyzed": "/Engine/Deck2/Track/SongAnalyzed",
    "EngineDeck2TrackSongLoaded": "/Engine/Deck2/Track/SongLoaded",
    "EngineDeck2TrackSongName": "/Engine/Deck2/Track/SongName",
    "EngineDeck2TrackSoundSwitchGUID": "/Engine/Deck2/Track/SoundSwitchGuid",
    "EngineDeck2TrackTrackBytes": "/Engine/Deck2/Track/TrackBytes",
    "EngineDeck2TrackTrackData": "/Engine/Deck2/Track/TrackData",
    "EngineDeck2TrackTrackLength": "/Engine/Deck2/Track/TrackLength",
    "EngineDeck2TrackTrackName": "/Engine/Deck2/Track/TrackName",
    "EngineDeck2TrackTrackNetworkPath": "/Engine/Deck2/Track/TrackNetworkPath",
    "EngineDeck2TrackTrackURI": "/Engine/Deck2/Track/TrackUri",
    "EngineDeck2TrackTrackWasPlayed": "/Engine/Deck2/Track/TrackWasPlayed",
    "EngineDeck3CurrentBPM": "/Engine/Deck3/CurrentBPM",
    "EngineDeck3ExternalMixerVolume": "/Engine/Deck3/ExternalMixerVolume",
    "EngineDeck3ExternalScratchWheelTouch": "/Engine/Deck3/ExternalScratchWheelTouch",
    "EngineDeck3PadsView": "/Engine/Deck3/Pads/View",
    "EngineDeck3Play": "/Engine/Deck3/Play",
    "EngineDeck3PlayState": "/Engine/Deck3/PlayState",
    "EngineDeck3PlayStatePath": "/Engine/Deck3/PlayStatePath",
    "EngineDeck3Speed": "/Engine/Deck3/Speed",
    "EngineDeck3SpeedNeutral": "/Engine/Deck3/SpeedNeutral",
    "EngineDeck3SpeedOffsetDown": "/Engine/Deck3/SpeedOffsetDown",
    "EngineDeck3SpeedOffsetUp": "/Engine/Deck3/SpeedOffsetUp",
    "EngineDeck3SpeedRange": "/Engine/Deck3/SpeedRange",
    "EngineDeck3SpeedState": "/Engine/Deck3/SpeedState",
    "EngineDeck3SyncMode": "/Engine/Deck3/SyncMode",
    "EngineDeck3TrackArtistName": "/Engine/Deck3/Track/ArtistName",
    "EngineDeck3TrackBleep": "/Engine/Deck3/Track/Bleep",
    "EngineDeck3TrackCuePosition": "/Engine/Deck3/Track/CuePosition",
    "EngineDeck3TrackCurrentBPM": "/Engine/Deck3/Track/CurrentBPM",
    "EngineDeck3TrackCurrentKeyIndex": "/Engine/Deck3/Track/CurrentKeyIndex",
    "EngineDeck3TrackCurrentLoopInPosition": "/Engine/Deck3/Track/CurrentLoopInPosition",
    "EngineDeck3TrackCurrentLoopOutPosition": "/Engine/Deck3/Track/CurrentLoopOutPosition",
    "EngineDeck3TrackCurrentLoopSizeInBeats": "/Engine/Deck3/Track/CurrentLoopSizeInBeats",
    "EngineDeck3TrackKeyLock": "/Engine/Deck3/Track/KeyLock",
    "EngineDeck3TrackLoopEnableState": "/Engine/Deck3/Track/LoopEnableState",
    "EngineDeck3TrackLoopQuickLoop1": "/Engine/Deck3/Track/Loop/QuickLoop1",
    "EngineDeck3TrackLoopQuickLoop2": "/Engine/Deck3/Track/Loop/QuickLoop2",
    "EngineDeck3TrackLoopQuickLoop3": "/Engine/Deck3/Track/Loop/QuickLoop3",
    "EngineDeck3TrackLoopQuickLoop4": "/Engine/Deck3/Track/Loop/QuickLoop4",
    "EngineDeck3TrackLoopQuickLoop5": "/Engine/Deck3/Track/Loop/QuickLoop5",
    "EngineDeck3TrackLoopQuickLoop6": "/Engine/Deck3/Track/Loop/QuickLoop6",
    "EngineDeck3TrackLoopQuickLoop7": "/Engine/Deck3/Track/Loop/QuickLoop7",
    "EngineDeck3TrackLoopQuickLoop8": "/Engine/Deck3/Track/Loop/QuickLoop8",
    "EngineDeck3TrackPlayPauseLEDState": "/Engine/Deck3/Track/PlayPauseLEDState",
    "EngineDeck3TrackSampleRate": "/Engine/Deck3/Track/SampleRate",
    "EngineDeck3TrackSongAnalyzed": "/Engine/Deck3/Track/SongAnalyzed",
    "EngineDeck3TrackSongLoaded": "/Engine/Deck3/Track/SongLoaded",
    "EngineDeck3TrackSongName": "/Engine/Deck3/Track/SongName",
    "EngineDeck3TrackSoundSwitchGUID": "/Engine/Deck3/Track/SoundSwitchGuid",
    "EngineDeck3TrackTrackBytes": "/Engine/Deck3/Track/TrackBytes",
    "EngineDeck3TrackTrackData": "/Engine/Deck3/Track/TrackData",
    "EngineDeck3TrackTrackLength": "/Engine/Deck3/Track/TrackLength",
    "EngineDeck3TrackTrackName": "/Engine/Deck3/Track/TrackName",
    "EngineDeck3TrackTrackNetworkPath": "/Engine/Deck3/Track/TrackNetworkPath",
    "EngineDeck3TrackTrackURI": "/Engine/Deck3/Track/TrackUri",
    "EngineDeck3TrackTrackWasPlayed": "/Engine/Deck3/Track/TrackWasPlayed",
    "EngineDeck4CurrentBPM": "/Engine/Deck4/CurrentBPM",
    "EngineDeck4ExternalMixerVolume": "/Engine/Deck4/ExternalMixerVolume",
    "EngineDeck4ExternalScratchWheelTouch": "/Engine/Deck4/ExternalScratchWheelTouch",
    "EngineDeck4PadsView": "/Engine/Deck4/Pads/View",
    "EngineDeck4Play": "/Engine/Deck4/Play",
    "EngineDeck4PlayState": "/Engine/Deck4/PlayState",
    "EngineDeck4PlayStatePath": "/Engine/Deck4/PlayStatePath",
    "EngineDeck4Speed": "/Engine/Deck4/Speed",
    "EngineDeck4SpeedNeutral": "/Engine/Deck4/SpeedNeutral",
    "EngineDeck4SpeedOffsetDown": "/Engine/Deck4/SpeedOffsetDown",
    "EngineDeck4SpeedOffsetUp": "/Engine/Deck4/SpeedOffsetUp",
    "EngineDeck4SpeedRange": "/Engine/Deck4/SpeedRange",
    "EngineDeck4SpeedState": "/Engine/Deck4/SpeedState",
    "EngineDeck4SyncMode": "/Engine/Deck4/SyncMode",
    "EngineDeck4TrackArtistName": "/Engine/Deck4/Track/ArtistName",
    "EngineDeck4TrackBleep": "/Engine/Deck4/Track/Bleep",
    "EngineDeck4TrackCuePosition": "/Engine/Deck4/Track/CuePosition",
    "EngineDeck4TrackCurrentBPM": "/Engine/Deck4/Track/CurrentBPM",
    "EngineDeck4TrackCurrentKeyIndex": "/Engine/Deck4/Track/CurrentKeyIndex",
    "EngineDeck4TrackCurrentLoopInPosition": "/Engine/Deck4/Track/CurrentLoopInPosition",
    "EngineDeck4TrackCurrentLoopOutPosition": "/Engine/Deck4/Track/CurrentLoopOutPosition",
    "EngineDeck4TrackCurrentLoopSizeInBeats": "/Engine/Deck4/Track/CurrentLoopSizeInBeats",
    "EngineDeck4TrackKeyLock": "/Engine/Deck4/Track/KeyLock",
    "EngineDeck4TrackLoopEnableState": "/Engine/Deck4/Track/LoopEnableState",
    "EngineDeck4TrackLoopQuickLoop1": "/Engine/Deck4/Track/Loop/QuickLoop1",
    "EngineDeck4TrackLoopQuickLoop2": "/Engine/Deck4/Track/Loop/QuickLoop2",
    "EngineDeck4TrackLoopQuickLoop3": "/Engine/Deck4/Track/Loop/QuickLoop3",
    "EngineDeck4TrackLoopQuickLoop4": "/Engine/Deck4/Track/Loop/QuickLoop4",
    "EngineDeck4TrackLoopQuickLoop5": "/Engine/Deck4/Track/Loop/QuickLoop5",
    "EngineDeck4TrackLoopQuickLoop6": "/Engine/Deck4/Track/Loop/QuickLoop6",
    "EngineDeck4TrackLoopQuickLoop7": "/Engine/Deck4/Track/Loop/QuickLoop7",
    "EngineDeck4TrackLoopQuickLoop8": "/Engine/Deck4/Track/Loop/QuickLoop8",
    "EngineDeck4TrackPlayPauseLEDState": "/Engine/Deck4/Track/PlayPauseLEDState",
    "EngineDeck4TrackSampleRate": "/Engine/Deck4/Track/SampleRate",
    "EngineDeck4TrackSongAnalyzed": "/Engine/Deck4/Track/SongAnalyzed",
    "EngineDeck4TrackSongLoaded": "/Engine/Deck4/Track/SongLoaded",
    "EngineDeck4TrackSongName": "/Engine/Deck4/Track/SongName",
    "EngineDeck4TrackSoundSwitchGUID": "/Engine/Deck4/Track/SoundSwitchGuid",
    "EngineDeck4TrackTrackBytes": "/Engine/Deck4/Track/TrackBytes",
    "EngineDeck4TrackTrackData": "/Engine/Deck4/Track/TrackData",
    "EngineDeck4TrackTrackLength": "/Engine/Deck4/Track/TrackLength",
    "EngineDeck4TrackTrackName": "/Engine/Deck4/Track/TrackName",
    "EngineDeck4TrackTrackNetworkPath": "/Engine/Deck4/Track/TrackNetworkPath",
    "EngineDeck4TrackTrackURI": "/Engine/Deck4/Track/TrackUri",
    "EngineDeck4TrackTrackWasPlayed": "/Engine/Deck4/Track/TrackWasPlayed",
    "EngineDeckCount": "/Engine/DeckCount",
    "GUIDecksDeckActiveDeck": "/GUI/Decks/Deck/ActiveDeck",
    "GUIViewLayerLayerB": "/GUI/ViewLayer/LayerB",
    "MixerCH1faderPosition": "/Mixer/CH1faderPosition",
    "MixerCH2faderPosition": "/Mixer/CH2faderPosition",
    "MixerCH3faderPosition": "/Mixer/CH3faderPosition",
    "MixerCH4faderPosition": "/Mixer/CH4faderPosition",
    "MixerCrossfaderPosition": "/Mixer/CrossfaderPosition",
}


@dataclass
class ServiceHandle:
    device: str
    ip: str
    service: str
    port: int


class StateMapSubscription:
    def __init__(
        self,
        service_handle: ServiceHandle,
        subscription_list: dict,
        data_callback: Callable[[list[DataClasses.StageLinQStateMapData]], None],
    ) -> None:
        self.service_handle: ServiceHandle = service_handle
        self._callback = data_callback
        self._subscription_list = subscription_list
        self.reader = None
        self.writer = None
        self.state_map_task = None

    def get_task(self) -> asyncio.Task:
        return self.state_map_task

    async def read_state_map(self) -> None:
        trailing_data = bytearray()
        while True:
            frame = await self.read_frame(trailing_data)

            if len(frame) == 0:
                return

            blocks, trailing_data = self.decode_frame(frame)

            self.handle_data(blocks)

    def decode_frame(self, frame: bytes) -> [list[bytes], bytearray]:
        blocks = self.decode_multi_block(frame)
        trailing_data = bytearray()
        if len(blocks[-1]) < 4:
            trailing_data = blocks.pop()
        elif len(blocks[-1]) != int.from_bytes(blocks[-1][0:4], byteorder="big") + 4:
            # Last block not completely received
            trailing_data = blocks.pop()
        return blocks, trailing_data

    def handle_data(self, blocks: list[bytes]) -> None:
        received_fields = []
        verify_error = None
        for block in blocks:
            try:
                received_fields.append(self.verify_block(block))

            except RuntimeError as err:
                # block corrupted, send error to callback if it exists
                received_fields = err
                verify_error = err
                break

        if self._callback is not None:
            self._callback(received_fields)
        else:
            # To avoid failing silently, the error is reraised if no callback has been defined.
            if verify_error is not None:
                raise verify_error from None

    async def read_frame(self, trailing_data) -> bytes:
        # Massive buffer because I have not figured out how to handle messages over multiple buffers
        frame = await self.reader.read(8192 * 4)
        # Add data from last received frame
        frame = trailing_data + frame
        return frame

    @staticmethod
    def decode_multi_block(frame: bytes) -> list[bytes]:
        blocks = []
        while len(frame) > 4:
            block_len = int.from_bytes(frame[0:4], byteorder="big") + 4
            blocks.append(frame[0:block_len])
            frame = frame[block_len:]

        if len(frame) > 0:
            # Frame not completely received, appending trailing block
            blocks.append(frame)

        return blocks

    @staticmethod
    def verify_block(block: bytes) -> DataClasses.StageLinQStateMapData:
        if len(block) < 4:
            raise RuntimeError("Block is to short to contain length")

        block_length = int.from_bytes(block[0:4], byteorder="big")

        if block_length == 0:
            # No data in the block, this most likely is due to other problem, or will cause problems later.
            return

        # Note: -4 is needed as block_length does not include its own length.
        if len(block) - 4 != block_length:
            raise RuntimeError(
                "Block invalid: Block length inconsistent with its header"
            )

        magic_flag = block[4:8].decode()

        if magic_flag != "smaa":
            raise RuntimeError("Block invalid: Could not find magic flag")

        # magic_flag2 is not checked as I have no idea what it is, it is all 0's for my PrimeGo, but it seems to be
        # different in the Go implementation
        magic_flag2 = block[8:12]

        path_len = int.from_bytes(block[12:16], byteorder="big")
        value_len_offset = 16 + path_len
        path = block[16:value_len_offset].decode(encoding="UTF-16be")

        value_offset = value_len_offset + 4
        value_len = int.from_bytes(
            block[value_len_offset:value_offset], byteorder="big"
        )

        value = json.loads(
            block[value_offset : value_offset + value_len].decode(encoding="UTF-16be")
        )

        return DataClasses.StageLinQStateMapData(
            block_length, magic_flag, magic_flag2, path_len, path, value_len, value
        )

    async def subscribe(self, own_token: Token.StageLinQToken) -> None:
        self.reader, self.writer = await asyncio.open_connection(
            self.service_handle.ip, self.service_handle.port
        )
        self.writer.write(
            StageLinQServiceAnnouncement().encode_frame(
                StageLinQServiceAnnouncementData(
                    Token=own_token,
                    Service=self.service_handle.service,
                    Port=self.writer.transport.get_extra_info("sockname")[1],
                )
            )
        )
        await self.writer.drain()

        self.state_map_task = asyncio.create_task(self.read_state_map())

        await self._send_subscription_requests()

    async def _send_subscription_requests(self):
        for service in self._subscription_list.values():
            msg = (len(service) * 2 + 8 * 2).to_bytes(4, byteorder="big")
            msg += bytearray([0x73, 0x6D, 0x61, 0x61])
            msg += bytearray([0x00, 0x00, 0x07, 0xD2])
            msg += StageLinQMessage().write_network_string(service)
            msg += (0).to_bytes(4, byteorder="big")
            self.writer.write(msg)
            await self.writer.drain()
        # small sleep to actually send frame
        await asyncio.sleep(0.001)



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""


class PyStageLinQError:
    # PyStageLinQ Error codes
    STAGELINQOK = 0

    DISCOVERYTIMEOUT = 100

    INVALIDFRAME = 200
    MAGICFLAGNOTFOUND = 201
    INVALIDLENGTH = 202
    BLOCKLENGTHINCONSISTENT = 203

    INVALIDTOKEN = 300
    INVALIDDISCOVERYDATA = 301
    INVALIDTOKENTYPE = 302
    INVALIDTOKENLENGTH = 303
    INVALIDTOKENVALUE = 304

    CANNOTBINDSOCKET = 400

    SERVICENOTRECOGNIZED = 500



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

from .Token import *
from dataclasses import dataclass


class StageLinQMessageIDs:
    StageLinQServiceAnnouncementData = (0).to_bytes(4, byteorder="big")
    StageLinQReferenceData = (1).to_bytes(4, byteorder="big")
    StageLinQServiceRequestData = (2).to_bytes(4, byteorder="big")


@dataclass
class StageLinQDiscoveryData:
    Token: StageLinQToken
    DeviceName: str
    ConnectionType: str
    SwName: str
    SwVersion: str
    ReqServicePort: int


@dataclass
class StageLinQServiceAnnouncementData:
    Token: StageLinQToken
    Service: str
    Port: int


@dataclass
class StageLinQReferenceData:
    OwnToken: StageLinQToken
    DeviceToken: StageLinQToken
    Reference: int


@dataclass
class StageLinQServiceRequestService:
    Token: StageLinQToken


@dataclass
class StageLinQStateMapData:
    BlockLength: int
    MagicFlag: str
    MagicFlag2: str
    ParameterLength: int
    ParameterName: str
    ValueLength: int
    ParameterValue: str



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

from .DataClasses import *


class ConnectionTypes:
    HOWDY = "DISCOVERER_HOWDY_"
    EXIT = "DISCOVERER_EXIT_"


class StageLinQMessage:
    port_length = 2
    magic_flag_start = 0
    magic_flag_length = 4
    magic_flag_stop = magic_flag_start + magic_flag_length

    network_len_size = 4
    reference_len = 8

    length: int | type(None)

    def write_network_string(self, string: str) -> bytes:
        # Note: 2 bytes are needed for every char as it is UTF16 encoded
        ret = (2 * len(string)).to_bytes(self.network_len_size, byteorder="big")
        ret += string.encode(encoding="UTF-16be")

        return ret

    def read_network_string(self, frame, start_offset):
        # Only uint32 length supported
        size_stop = start_offset + self.network_len_size
        data_start = size_stop

        data_length = int.from_bytes(frame[start_offset:size_stop], byteorder="big")
        data_stop = data_start + data_length

        if data_stop > len(frame):
            # Out of bounds
            raise Exception(PyStageLinQError.INVALIDLENGTH)

        return data_stop, frame[data_start:data_stop].decode(encoding="UTF-16be")

    def get_len(self):
        return self.length


class StageLinQDiscovery(StageLinQMessage):
    token: StageLinQToken
    deviceName: str
    connectionType: str
    sw_name: str | type(None)
    sw_version: str | type(None)
    stagelinq_magic_flag = bytes("airD".encode(encoding="ASCII"))

    def __init__(self):
        self.Port = None
        self.sw_version = None
        self.sw_name = None
        self.device_name = None
        self.connection_type = None
        self.token = StageLinQToken()
        self.length = 0
        self.min_length = (
            self.magic_flag_length + StageLinQToken.TOKENLENGTH + self.network_len_size
        )

    def encode_frame(self, discovery_data: StageLinQDiscoveryData) -> bytes:
        if self.verify_discovery_data(discovery_data) != PyStageLinQError.STAGELINQOK:
            return PyStageLinQError.INVALIDDISCOVERYDATA

        discovery_frame = self.stagelinq_magic_flag
        discovery_frame += discovery_data.Token.get_token().to_bytes(
            StageLinQToken.TOKENLENGTH, byteorder="big"
        )
        discovery_frame += self.write_network_string(discovery_data.DeviceName)
        discovery_frame += self.write_network_string(discovery_data.ConnectionType)
        discovery_frame += self.write_network_string(discovery_data.SwName)
        discovery_frame += self.write_network_string(discovery_data.SwVersion)
        discovery_frame += discovery_data.ReqServicePort.to_bytes(
            self.port_length, byteorder="big"
        )

        return discovery_frame

    @staticmethod
    def verify_discovery_data(discovery_data) -> StageLinQDiscoveryData:
        ret = PyStageLinQError.STAGELINQOK
        if (0 > discovery_data.ReqServicePort) or (
            65535 < discovery_data.ReqServicePort
        ):
            ret = PyStageLinQError.INVALIDDISCOVERYDATA
        return ret

    def get(self):
        return StageLinQDiscoveryData(
            self.token,
            self.device_name,
            self.connection_type,
            self.sw_name,
            self.sw_version,
            self.Port,
        )

    def decode_frame(self, frame):
        if len(frame) < self.min_length:
            return PyStageLinQError.INVALIDFRAME

        # Local Constants
        token_start = self.magic_flag_stop
        token_length = StageLinQToken.TOKENLENGTH
        token_stop = token_start + token_length
        device_name_size_start = token_stop

        # Check if Frame contains Magic flag
        if (
            self.stagelinq_magic_flag
            != frame[self.magic_flag_start : self.magic_flag_stop]
        ):
            return PyStageLinQError.MAGICFLAGNOTFOUND

        self.token.set_token(
            int.from_bytes(frame[token_start:token_stop], byteorder="big")
        )

        try:
            connection_type_start, self.device_name = self.read_network_string(
                frame, device_name_size_start
            )

            sw_name_start, self.connection_type = self.read_network_string(
                frame, connection_type_start
            )

            sw_version_start, self.sw_name = self.read_network_string(
                frame, sw_name_start
            )
            port_start, self.sw_version = self.read_network_string(
                frame, sw_version_start
            )

        except Exception as e:
            if str(e) == str(PyStageLinQError.INVALIDLENGTH):
                return PyStageLinQError.INVALIDLENGTH
            else:
                raise e

        port_stop = port_start + self.port_length

        if len(frame) < port_stop:
            return PyStageLinQError.INVALIDLENGTH

        self.Port = int.from_bytes(frame[port_start:port_stop], byteorder="big")
        self.length = port_stop
        return PyStageLinQError.STAGELINQOK


class StageLinQServiceAnnouncement(StageLinQMessage):
    Token: StageLinQToken | type(None)
    Service: str | type(None)
    Port: int | type(None)

    def __init__(self):
        self.Token = StageLinQToken()
        self.Service = None
        self.Port = None
        self.length = None

        self.min_length = (
            self.magic_flag_length + StageLinQToken.TOKENLENGTH + self.network_len_size
        )

    def encode_frame(
        self, service_announcement_data: StageLinQServiceAnnouncementData
    ) -> bytes:
        request_frame = StageLinQMessageIDs.StageLinQServiceAnnouncementData
        request_frame += service_announcement_data.Token.get_token().to_bytes(
            StageLinQToken.TOKENLENGTH, byteorder="big"
        )
        request_frame += self.write_network_string(service_announcement_data.Service)
        request_frame += service_announcement_data.Port.to_bytes(2, byteorder="big")
        return request_frame

    def decode_frame(self, frame):
        if len(frame) < self.min_length:
            return PyStageLinQError.INVALIDLENGTH

        # Verify frame type
        if (
            frame[self.magic_flag_start : self.magic_flag_stop]
            != StageLinQMessageIDs.StageLinQServiceAnnouncementData
        ):
            return PyStageLinQError.INVALIDFRAME

        token_start = self.magic_flag_stop
        token_stop = token_start + StageLinQToken.TOKENLENGTH
        service_name_start = token_stop

        self.Token.set_token(
            (0).from_bytes(frame[token_start:token_stop], byteorder="big")
        )

        try:
            port_start, self.Service = self.read_network_string(
                frame, service_name_start
            )
        except Exception as e:
            if str(e) == str(PyStageLinQError.INVALIDLENGTH):
                return PyStageLinQError.INVALIDLENGTH
            else:
                raise e

        port_stop = port_start + self.port_length

        if len(frame) < port_stop:
            return PyStageLinQError.INVALIDLENGTH

        self.Port = int.from_bytes(frame[port_start:port_stop], byteorder="big")
        self.length = port_stop
        return PyStageLinQError.STAGELINQOK

    def get(self):
        return StageLinQServiceAnnouncementData(
            Token=self.Token, Service=self.Service, Port=self.Port
        )


class StageLinQReference(StageLinQMessage):
    OwnToken: StageLinQToken | type(None)
    DeviceToken: StageLinQToken | type(None)
    Reference: int | type(None)

    def __init__(self):
        self.OwnToken = StageLinQToken()
        self.DeviceToken = StageLinQToken()
        self.Reference = None
        self.length = (
            self.magic_flag_length + StageLinQToken.TOKENLENGTH * 2 + self.reference_len
        )

    @staticmethod
    def encode_frame(reference_data) -> StageLinQReferenceData:
        request_frame = StageLinQMessageIDs.StageLinQReferenceData
        request_frame += reference_data.OwnToken.get_token().to_bytes(
            StageLinQToken.TOKENLENGTH, byteorder="big"
        )
        request_frame += 0x00.to_bytes(StageLinQToken.TOKENLENGTH, byteorder="big")
        request_frame += reference_data.Reference.to_bytes(8, byteorder="big")
        return request_frame

    def decode_frame(self, frame):
        if len(frame) < self.length:
            return PyStageLinQError.INVALIDLENGTH

        # Verify frame type
        if (
            frame[self.magic_flag_start : self.magic_flag_stop]
            != StageLinQMessageIDs.StageLinQReferenceData
        ):
            return PyStageLinQError.INVALIDFRAME

        own_token_start = self.magic_flag_stop
        own_token_stop = own_token_start + StageLinQToken.TOKENLENGTH
        device_token_start = own_token_stop
        device_token_stop = device_token_start + StageLinQToken.TOKENLENGTH
        reference_start = device_token_stop
        reference_stop = reference_start + self.reference_len

        self.OwnToken.set_token(
            (0).from_bytes(frame[own_token_start:own_token_stop], byteorder="big")
        )

        token_id = (0).from_bytes(
            frame[device_token_start:device_token_stop], byteorder="big"
        )

        if token_id != 0:
            self.DeviceToken.set_token(token_id)
        self.Reference = int.from_bytes(
            frame[reference_start:reference_stop], byteorder="big"
        )
        return PyStageLinQError.STAGELINQOK

    def get(self):
        return StageLinQReferenceData(
            OwnToken=self.OwnToken,
            DeviceToken=self.DeviceToken,
            Reference=self.Reference,
        )


class StageLinQRequestServices(StageLinQMessage):
    Token: StageLinQToken

    length = 4 + StageLinQToken.TOKENLENGTH

    def __init__(self):
        self.Token = StageLinQToken()

    @staticmethod
    def encode_frame(service_request_data) -> StageLinQServiceRequestService:
        request_frame = StageLinQMessageIDs.StageLinQServiceRequestData
        request_frame += service_request_data.Token.get_token().to_bytes(
            StageLinQToken.TOKENLENGTH, byteorder="big"
        )
        return request_frame

    def decode_frame(self, frame):
        # Verify frame
        if len(frame) < self.length:
            return PyStageLinQError.INVALIDLENGTH
        if (
            frame[self.magic_flag_start : self.magic_flag_stop]
            != StageLinQMessageIDs.StageLinQServiceRequestData
        ):
            return PyStageLinQError.INVALIDFRAME

        token_start = self.magic_flag_stop
        token_stop = token_start + StageLinQToken.TOKENLENGTH

        self.Token = frame[token_start:token_stop]
        return PyStageLinQError.STAGELINQOK

    def get(self):
        return StageLinQServiceRequestService(Token=self.Token)



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

from random import randbytes
from .ErrorCodes import PyStageLinQError


class StageLinQToken:
    TOKENLENGTH = 16
    token: int

    def __init__(self):
        self.token = 0

    def generate_token(self):
        """
        A function to generate the token.

        NOTE:
        The tokens MSB cannot have the 0x80 bit set, this will cause the device to not send services

        :return: No return value
        """
        randomized_bytes = self._get_randomized_bytes(self.TOKENLENGTH)

        # The following checks are to solve issues with invalid token values I've found when testing. The actual token
        # structure is currently unknown, so a random value is generated as this seems to be what's going on from the
        # devices I've tested against.
        invalid_token_bit_mask = ~(0x1 << 127 | 0xC << 60)

        self.token = (
            int.from_bytes(randomized_bytes, byteorder="big") & invalid_token_bit_mask
        )

    @staticmethod
    def _get_randomized_bytes(length: int) -> bytes:
        randomized_bytes = bytearray(randbytes(length))
        return randomized_bytes

    def get_token(self) -> int:
        return self.token

    def set_token(self, token: int) -> None:
        if type(token) == int:
            if self.validate_token(token) == PyStageLinQError.STAGELINQOK:
                self.token = token
            else:
                # Token could not be Validated
                raise PyStageLinQError.INVALIDTOKEN
        else:
            # Token is not of type int
            raise PyStageLinQError.INVALIDTOKENTYPE

    @staticmethod
    def validate_token(token: int) -> PyStageLinQError:
        # The token is validated by converting it to a 16 byte array and then back to an int. If the value is the same
        # the token is considered valid
        if token.bit_length() <= StageLinQToken.TOKENLENGTH * 8:
            if token > 0:
                ret = PyStageLinQError.STAGELINQOK

            else:
                ret = PyStageLinQError.INVALIDTOKEN

        else:
            ret = PyStageLinQError.INVALIDTOKENLENGTH

        return ret



from . import DataClasses
from . import Network


class DeviceList:
    device_list: [DataClasses.StageLinQDiscoveryData]

    def __init__(self):
        self.device_list = []

    def register_device(self, device: Network.StageLinQService) -> bool:
        if type(device) is not Network.StageLinQService:
            return False
        self.device_list.append(device)
        return True

    def find_registered_device(
        self, discovery_frame: DataClasses.StageLinQDiscoveryData
    ) -> bool:
        # Check if main link has been established for Offline analyzers
        if discovery_frame.SwName == "OfflineAnalyzer":
            if not self.find_main_interface(discovery_frame):
                return True

        # Check if device is registered
        for device in self.device_list:
            if discovery_frame.DeviceName == device.device_name:
                if discovery_frame.ReqServicePort == device.Port:
                    # Device already registered
                    return True
        return False

    def find_main_interface(
        self, discovery_frame: DataClasses.StageLinQDiscoveryData
    ) -> bool:
        for device in self.device_list:
            if device.device_name == discovery_frame.DeviceName:
                if device.sw_name != "OfflineAnalyzer":
                    return True

        return False



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

import select
import socket
import time
import asyncio
import logging
import platform
import psutil
import ipaddress
from typing import Callable

from . import Device
from .MessageClasses import *
from .DataClasses import *
from .ErrorCodes import PyStageLinQError
from .Network import StageLinQService
from . import EngineServices

logger = logging.getLogger("PyStageLinQ")


@dataclass
class PyStageLinQ_interface_info:
    name: str
    id: int
    addr: int
    addr_str: str
    mask: int
    status: psutil._common.snicstats
    n_disc_msg_send: int


class PyStageLinQ_network_interface:
    def __init__(self, ip=None, discovery_port=51337):
        self.target_interfaces = []
        self.discovery_port = discovery_port
        self.get_interface_from_ip(ip)

    def get_interface_from_ip(self, ip):
        if ip is None:
            ip_list = ["any"]
        elif type(ip) is str:
            ip_list = [ip]
        elif type(ip) is list:
            for entry in ip:
                if type(entry) is not str:
                    raise TypeError
            ip_list = ip
        else:
            raise TypeError

        for interface in psutil.net_if_stats().items():
            for interface_info in psutil.net_if_addrs()[interface[0]]:
                # Only look for IPV4 binds
                if socket.AF_INET == interface_info.family and (
                    interface_info.address in ip_list or ip_list[0] == "any"
                ):
                    self.target_interfaces.append(
                        PyStageLinQ_interface_info(
                            interface[0],
                            len(self.target_interfaces),
                            int(ipaddress.IPv4Address(interface_info.address)),
                            interface_info.address,
                            int(ipaddress.IPv4Address(interface_info.netmask)),
                            interface[1],
                            0,
                        )
                    )

    def send_discovery_frame(self, discovery_frame):
        for interface in self.target_interfaces:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as discovery_socket:
                discovery_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                discovery_socket.bind((interface.addr_str, 0))
                try:
                    discovery_socket.sendto(
                        discovery_frame,
                        ("255.255.255.255", self.discovery_port),
                    )
                    self.target_interfaces[interface.id].n_disc_msg_send += 1
                except PermissionError:
                    logger.warning(
                        f"Cannot send message on interface {interface.name}, "
                        f"this error could be due to that there is no network interface set up with this IP range"
                    )
                    raise PermissionError

    def determine_interface_of_remote_ip(self, ip):
        for interface in self.target_interfaces:
            if (int(ipaddress.IPv4Address(ip)) & interface.mask) == (
                interface.mask & interface.addr
            ):
                return interface

        return None

    def send_desc_on_all_if(self):
        for interface in self.target_interfaces:
            # Wait until a few discovery frames have been sent to make sure the other devices have seen us. If they have
            # not and we are asking for services it will be denied.
            if interface.n_disc_msg_send < 3:
                return False

        return True


class PyStageLinQ:
    """
    The main object for PyStageLinQ. Use this object to first initialize and then start PyStageLinq

    :param new_device_found_callback: This callback is used to report back to the application when a StageLinQ
    device has been detected on the network. It is then up to the application to determine what device has been
    and how if it wants to connect to it.

    :param name: This is the name which PyStageLinQ will announce itself with on the StageLinQ protocol. If not
    set it defaults to "Hello StageLinQ World".

    :param ip: This is the ip of the interface you want to bind the sockets to, e.g. your local ethernet IP. If set to
    None all interfaces will be used.

    """

    REQUESTSERVICEPORT = 0  # If set to anything but 0 other StageLinQ devices will try to request services at said port
    StageLinQ_discovery_port = 51337

    _loopcondition = True

    def __init__(
        self,
        new_device_found_callback: Callable[
            [str, StageLinQDiscovery, EngineServices.ServiceHandle], None
        ],
        name: str = "Hello StageLinQ World",
        ip=None,
    ):
        self.name = name
        self.OwnToken = StageLinQToken()
        self.OwnToken.generate_token()
        self.discovery_info = StageLinQDiscoveryData(
            Token=self.OwnToken,
            DeviceName=self.name,
            ConnectionType=ConnectionTypes.HOWDY,
            SwName="Python",
            SwVersion="0.2.2.dev5",
            ReqServicePort=self.REQUESTSERVICEPORT,
        )

        self.device_list = Device.DeviceList()

        self.network_interface = PyStageLinQ_network_interface(ip)

        self.tasks = set()

        self.active_services = []

        self.new_device_found_callback = new_device_found_callback

        logger.debug(f"Initialized!")

    def start_standalone(self):
        """
        Function for starting PyStageLinq in standalone mode. In this mode this function will not return, but
        instead start asyncio and que up tasks. This function shall be called once the PyStageLinQ object has been
        initialized.
        """
        logger.debug(f"Starting in standalone mode.")
        asyncio.run(self._start_stagelinq(standalone=True))

    def start(self):
        """
        Function for starting PyStageLinq as part of a bigger asyncio program. The function will start the required
        tasks and then return to its caller. It is then up to the called to make sure that the asyncio tasks gets
        time to execute. This function shall be called once the PyStageLinQ object has been initialized.
        """
        logger.debug(f"Starting in asyncio mode.")
        self._start_stagelinq()

    def _stop(self):
        logger.info(f"Stop requested, trying graceful shutdown")
        try:
            discovery = StageLinQDiscovery()
            discovery_info = self.discovery_info
            discovery_info.ConnectionType = ConnectionTypes.EXIT
            discovery_frame = discovery.encode_frame(discovery_info)

            self.network_interface.send_discovery_frame(discovery_frame)
            logger.info(f"Gracefully shutdown complete")
        except Exception as e:
            logger.debug('Could not send "EXIT" discovery frame during shutdown')
            raise e

    def _announce_self(self):
        discovery = StageLinQDiscovery()
        discovery_frame = discovery.encode_frame(self.discovery_info)
        self.network_interface.send_discovery_frame(discovery_frame)

    def get_loop_condition(self) -> bool:
        return self._loopcondition

    async def _discover_stagelinq_device(self, host_ip, timeout=10):
        """
        This function is used to find StageLinQ device announcements.
        """

        logger.info(f"Trying to discover StageLinQ devices.")

        # Local Constants
        discover_buffer_size = 8192

        # Create socket
        discover_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        try:
            discover_socket.bind(
                (host_ip, self.StageLinQ_discovery_port)
            )  # bind socket to broadcast
        except Exception as e:
            # Cannot bind to socket, check if IP is correct and link is up
            logger.warning(
                f"Cannot bind to IP socket: {host_ip} on port {self.StageLinQ_discovery_port}"
            )
            raise e
        discover_socket.setblocking(False)

        loop_timeout = time.time() + timeout

        logger.debug(
            f"Socket bound to IP {host_ip} and Port {self.StageLinQ_discovery_port} successfully. Starting to look for "
            f"discovery frames"
        )

        while self.get_loop_condition():
            data_available = select.select([discover_socket], [], [], 0)
            if data_available[0]:
                data, addr = discover_socket.recvfrom(discover_buffer_size)
                device_ip = addr[0]
                discovery_frame = StageLinQDiscovery()

                if PyStageLinQError.STAGELINQOK != discovery_frame.decode_frame(data):
                    # something went wrong
                    continue

                if self.name == discovery_frame.device_name:
                    # Ourselves, ignore message
                    continue

                if 0 == discovery_frame.Port:
                    # If port is 0 there are no services to request, and for our use an invalid device
                    continue

                # External Device present, setting new timeout
                loop_timeout = time.time() + timeout

                # Check if we have already registered this device, and if so ignore it for now
                device_registered = self.device_list.find_registered_device(
                    discovery_frame.get()
                )
                if device_registered is True:
                    continue

                logger.info(f"Found new StageLinq device at IP {device_ip}")
                await self._register_new_device(discovery_frame, device_ip)

            if time.time() > loop_timeout:
                # No devices found within timeout
                logger.info(
                    f"No discovery frames found on {host_ip} last {timeout} seconds."
                )
                return PyStageLinQError.DISCOVERYTIMEOUT
            await asyncio.sleep(0.1)

    async def _register_new_device(self, discovery_frame, ip):
        stagelinq_device = StageLinQService(ip, discovery_frame, self.OwnToken, None)
        service_tasks = await stagelinq_device.get_tasks()
        for task in service_tasks:
            self.tasks.add(task)
        self.device_list.register_device(stagelinq_device)
        logging.debug(f"Device found! Name: {stagelinq_device.device_name}")
        await stagelinq_device.wait_for_services(timeout=1)
        if self.new_device_found_callback is not None:
            self.new_device_found_callback(
                ip, discovery_frame, stagelinq_device.get_services()
            )

    def subscribe_to_statemap(
        self,
        state_map_service: EngineServices.ServiceHandle,
        subscription_list: dict,
        data_available_callback: Callable[[list[StageLinQStateMapData]], None] = None,
    ) -> None:
        """
        This function is used to subscribe to a statemap service provided by a StageLinQ device.
                :param state_map_service: This parameter is used to determine if
                :param subscription_list: list of serivces that the application wants to subscribe to
                :param data_available_callback: Callback for when data is available from StageLinQ device
        """
        if state_map_service.service != "StateMap":
            logger.warning(f'{state_map_service.service} is not of type "Statemap"')
            return PyStageLinQError.SERVICENOTRECOGNIZED

        # Defer task creation to avoid blocking the calling function
        asyncio.create_task(
            self._subscribe_to_statemap(
                state_map_service, subscription_list, data_available_callback
            )
        )
        return PyStageLinQError.STAGELINQOK

    async def _subscribe_to_statemap(
        self, state_map_service, subscription_list, data_available_callback
    ):
        state_map = EngineServices.StateMapSubscription(
            state_map_service, subscription_list, data_available_callback
        )
        await state_map.subscribe(self.OwnToken)

        self.tasks.add(state_map.get_task())
        logger.debug(f"Subscription to StateMap successful.")

    async def _start_stagelinq(self, standalone=False):
        # Start the initial tasks of the library
        self.tasks.add(asyncio.create_task(self._periodic_announcement()))

        # wait for discovery message to be send on all interfaces
        while not self.network_interface.send_desc_on_all_if():
            await asyncio.sleep(0.1)

        self.tasks.add(asyncio.create_task(self._py_stagelinq_strapper()))

        if standalone:
            await self._wait_for_exit()

    async def _wait_for_exit(self):
        while self.get_loop_condition():
            for task in self.tasks.copy():
                if task.done():
                    if task.exception() is not None:
                        logger.warning(
                            f"Exception: {task.exception()} occured in task: {task.get_coro()}, stopping "
                            f"PyStageLinQ."
                        )
                        logger.debug(f"Stacktrace to error: {task.print_stack()}")
                        self.stop()
                        raise task.exception()
                    return
            await asyncio.sleep(1)

    def _stop_all_tasks(self):
        logger.info(f"Stop all PyStageLinQ AsyncIO tasks requested")
        for task in self.tasks.copy():
            task.cancel()

    async def _periodic_announcement(self):
        while self.get_loop_condition():
            self._announce_self()
            await asyncio.sleep(0.5)

    async def _py_stagelinq_strapper(self):
        strapper_tasks = set()
        logger.info(f"Starting to look for StageLinQ discovery frames:")

        strapper_tasks.add(
            asyncio.create_task(self._discover_stagelinq_device("", timeout=2))
        )

        while self.get_loop_condition():
            all_tasks_done = True
            for task in strapper_tasks.copy():
                all_tasks_done = all_tasks_done and task.done()
                if task.done():
                    if task.exception() is not None:
                        raise task.exception()

            if all_tasks_done:
                logger.debug("Timeout occurred on all interfaces.")
                return
            await asyncio.sleep(1)

    def stop(self):
        self._stop()

    def __del__(self):
        self._stop_all_tasks()
        self._stop()



"""
(c) 2022 Jaxcie
This code is licensed under MIT license (see LICENSE for details)
"""

from __future__ import annotations
import asyncio
import logging
from . import EngineServices
from .DataClasses import (
    StageLinQServiceAnnouncementData,
    StageLinQReferenceData,
    StageLinQServiceRequestService,
)
from .MessageClasses import *
from . import Token
from typing import Callable, Tuple, List, Any

logger = logging.getLogger("PyStageLinQ")


class StageLinQService:
    _loopcondition = True

    def __init__(
        self,
        ip: str,
        discovery_frame: StageLinQDiscovery,
        own_token: StageLinQToken,
        service_found_callback: Callable[[StageLinQService], None],
    ) -> None:
        self.Socket = None
        self.reference_task = None
        self.reader = None
        self.writer = None
        self.Ip = ip
        self.Port = discovery_frame.get().ReqServicePort
        self.OwnToken = own_token
        self.DeviceToken = Token.StageLinQToken()
        self.service_list = []
        self.device_name = discovery_frame.device_name
        self.device_token = discovery_frame.token
        self.sw_name = discovery_frame.sw_name

        self.receive_task = asyncio.create_task(self.start_receive_data())

        self.init_complete = False
        self.services_available = False

        self.service_found_callback = service_found_callback

        self.remaining_data = None

        self.debug = []

    def get_services(self) -> list[EngineServices]:
        if self.services_available:
            return_list = []
            for service in self.service_list:
                return_list.append(
                    EngineServices.ServiceHandle(
                        device=self.device_name,
                        ip=self.Ip,
                        service=service[0],
                        port=service[1],
                    )
                )
            return return_list

    def get_init_complete(self) -> bool:
        return self.init_complete

    async def get_tasks(self) -> [asyncio.Task, asyncio.Task]:
        while self.get_init_complete() is False:
            await asyncio.sleep(0.1)
        return [self.reference_task, self.receive_task]

    def get_services_available(self) -> bool:
        return self.services_available

    async def wait_for_services(self, timeout: float = -1.0) -> None:
        timestep = 0.01
        n = 0
        while True:
            if self.get_services_available() is True:
                return
            elif timeout >= 0 and n > timeout / timestep:
                raise RuntimeError("Timeout occurred before services were received")
            n = n + 1
            await asyncio.sleep(timestep)

    async def send_request_frame(self) -> None:
        out_data = StageLinQRequestServices()
        service_request_message = out_data.encode_frame(
            StageLinQServiceRequestService(Token=self.OwnToken)
        )

        self.writer.write(service_request_message)
        await self.writer.drain()

    async def start_receive_data(self) -> None:
        # Send first message

        self.reader, self.writer = await asyncio.open_connection(self.Ip, self.Port)
        self.reference_task = asyncio.create_task(
            self.send_reference_message_periodically()
        )
        asyncio.create_task(self.send_request_frame())

        self.init_complete = True

        await self._receive_data_loop()

    def get_loop_condition(self) -> bool:
        return self._loopcondition

    async def _receive_data_loop(self) -> None:
        while self.get_loop_condition() is True:
            frames = await self._receive_frames()

            if frames is False:
                continue

            await self._handle_frames(frames)

    async def _receive_frames(
        self,
    ) -> (
        bool
        | list[
            StageLinQServiceAnnouncementData
            | StageLinQReferenceData
            | StageLinQServiceRequestService
        ]
    ):
        response = await self.reader.read(1024)
        if len(response) == 0:
            # Socket closed
            raise RuntimeError(
                f"Remote socket for IP:{self.Ip} Port:{self.Port} closed!"
            )

        if self.remaining_data is not None:
            response = b"".join([self.remaining_data, response])
        frames, self.remaining_data = self.decode_multiframe(response)
        if frames is None:
            # Something went wrong during decoding, lets throw away the frame and hope it doesn't happen again
            logger.debug(f"Error while decoding the frame")
            return False
        self.last_frame = response
        return frames

    async def _handle_frames(self, frames: bytes) -> None:
        adding_services = False
        for frame in frames:
            if type(frame) is StageLinQServiceAnnouncementData:
                self.service_list.append([frame.Service, frame.Port])
                adding_services = True
                self._set_device_token(frame)

            elif type(frame) is StageLinQServiceRequestService:
                self._set_device_token(frame)

            elif type(frame) is StageLinQReferenceData:
                # Do not send a new frame if read frame is from ourselves.
                if frame.OwnToken.get_token() != self.OwnToken.get_token():
                    asyncio.create_task(self.send_reference_message())

        if adding_services:
            await self._handle_new_services()

    async def _handle_new_services(self) -> None:
        self.services_available = True
        if self.service_found_callback is not None:
            await self.service_found_callback(self)

    def _set_device_token(
        self,
        data_class: StageLinQServiceAnnouncementData | StageLinQServiceRequestService,
    ) -> None:
        if (type(data_class) is StageLinQServiceAnnouncementData) or (
            type(data_class) is StageLinQServiceRequestService
        ):
            if self.DeviceToken.get_token() == 0:
                self.DeviceToken.set_token(
                    int.from_bytes(data_class.Token, byteorder="big")
                )

    async def send_reference_message_periodically(self) -> None:
        while self.get_loop_condition():
            # Checking services_available is a really simple way of knowing if a link has been established, and initial
            # handshake is done
            if self.services_available:
                await self.send_reference_message()
            await asyncio.sleep(0.250)

    async def send_reference_message(self) -> None:
        reference_data = StageLinQReferenceData(
            OwnToken=self.OwnToken, DeviceToken=self.DeviceToken, Reference=0
        )
        reference_message = StageLinQReference()
        reference_message.encode_frame(reference_data)

        self.writer.write(reference_message.encode_frame(reference_data))
        await self.writer.drain()

    @staticmethod
    def decode_multiframe(
        frame: bytes,
    ) -> tuple[list[Any], None | bytes] | None:
        subframes = []
        while len(frame) >= 4:
            match (int.from_bytes(frame[0:4], byteorder="big")):
                case 0:
                    data = StageLinQServiceAnnouncement()
                case 1:
                    data = StageLinQReference()
                case 2:
                    data = StageLinQRequestServices()
                case _:
                    # invalid data, return
                    return None
            decode_status = data.decode_frame(frame)

            if decode_status != PyStageLinQError.STAGELINQOK:
                if decode_status == PyStageLinQError.INVALIDLENGTH:
                    return subframes, frame
                else:
                    return None

            subframes.append(data.get())
            frame = frame[data.get_len() :]

        return subframes, None
